// copyright 2023 zamfofex
// license: AGPLv3 or later

#include <mimimi/allocators.h>
#include <mimimi/animations.h>
#include <mimimi/assets.h>
#include <mimimi/poses.h>
#include <mimimi/video.h>
#include <mimimi/sprites.h>
#include <mimimi/chapters.h>
#include <mimimi/geometry.h>
#include <mimimi/behaviors.h>
#include <mimimi/coroutines.h>
#include <mimimi/engines.h>

#include <mimimi/chapter-I/appearances.h>

static void mimimi_chapter_I_animation_set(struct mimimi_chapter_I_animation_set *set, struct mimimi_allocator *allocator)
{
	int animation_count = 8;
	int image_count = 8;
	int width = 48;
	int height = 48;
	
	int total_animation_count = 4 + animation_count;
	int total_image_count = 2 + animation_count * image_count + 2 * image_count;
	
	struct mimimi_animation *animations = mimimi_allocate(allocator, total_animation_count * sizeof *animations);
	struct mimimi_image *images = mimimi_allocate(allocator, total_image_count * sizeof *images);
	
	set->standing = animations;
	set->standing_animation_count = animation_count;
	animations += animation_count;
	
	for (int i = 0 ; i < animation_count ; i++)
	{
		struct mimimi_animation *animation = set->standing + i;
		animation->count = image_count;
		animation->images = images;
		images += image_count;
		
		for (int i = 0 ; i < image_count ; i++)
		{
			struct mimimi_image *image = animation->images + i;
			image->width = width;
			image->height = height;
			image->colors = mimimi_allocate(allocator, width * height);
			for (int i = 0 ; i < width * height ; i++) image->colors[i] = 0;
		}
	}
	
	struct mimimi_animation *knocked = animations++;
	knocked->count = 1;
	knocked->images = images++;
	knocked->images[0].width = width;
	knocked->images[0].height = height;
	knocked->images[0].colors = mimimi_allocate(allocator, width * height);
	for (int i = 0 ; i < width * height ; i++) knocked->images[0].colors[i] = 0;
	
	struct mimimi_animation *falling = animations++;
	falling->count = 1;
	falling->images = images++;
	falling->images[0].width = width;
	falling->images[0].height = height;
	falling->images[0].colors = mimimi_allocate(allocator, width * height);
	for (int i = 0 ; i < width * height ; i++) falling->images[0].colors[i] = 0;
	
	set->knocked = knocked;
	set->falling = falling;
	
	for (int i = 0 ; i < width * height ; i++)
	{
		knocked->images[0].colors[i] = 0;
		falling->images[0].colors[i] = 0;
	}
	
	set->jumping = animations++;
	set->jumping->count = image_count;
	set->jumping->images = images;
	images += image_count;
	
	set->landing = animations++;
	set->landing->count = image_count;
	set->landing->images = images;
	images += image_count;
	
	for (int i = 0 ; i < image_count ; i++)
	{
		struct mimimi_image *jumping = set->jumping->images + i;
		struct mimimi_image *landing = set->landing->images + i;
		
		jumping->width = width;
		jumping->height = height;
		jumping->colors = mimimi_allocate(allocator, width * height);
		
		landing->width = width;
		landing->height = height;
		landing->colors = mimimi_allocate(allocator, width * height);
		
		for (int i = 0 ; i < width * height ; i++)
			jumping->colors[i] = 0,
			landing->colors[i] = 0;
	}
}

static void mimimi_chapter_I_appearance_animations(struct mimimi_chapter_I_animation_set *animations, struct mimimi_model *model, int x, int y, int coefficient, struct mimimi_allocator *allocator)
{
	int animation_count = animations->standing_animation_count;
	for (int i = 0 ; i < animation_count ; i++)
	{
		struct mimimi_pose_layer layers[2][10] = {0};
		
		int slant = 512 * i / animation_count * coefficient;
		int y_angle = 64 + (20 + 20 * i / animation_count) * coefficient;
		int arm_slant = 64 * i / animation_count;
		
		layers[0][0].y_angle = y_angle;
		layers[1][0].y_angle = y_angle;
		
		layers[0][0].slant = -slant / 8;
		layers[1][0].slant = -slant / 8;
		
		layers[0][1].slant = slant / 8;
		layers[1][1].slant = slant / 8;
		
		layers[1][4].slant = slant / 3;
		layers[1][8].slant = -slant / 3;
		layers[0][8].slant = -slant / 8;
		
		layers[0][5].slant = slant / 3;
		layers[0][9].slant = -slant / 3;
		layers[1][9].slant = -slant / 8;
		
		layers[0][2].slant = slant / 5;
		layers[1][2].slant = -slant / 3;
		layers[0][6].slant = slant / 4;
		layers[1][6].slant = slant / 2;
		
		layers[1][3].slant = slant / 5;
		layers[0][3].slant = -slant / 3;
		layers[1][7].slant = slant / 4;
		layers[0][7].slant = slant / 2;
		
		int count = sizeof *layers / sizeof **layers;
		struct mimimi_pose poses[] = {{x, y, count, layers[0]}, {x, y, count, layers[1]}, {x, y, count, layers[0]}};
		struct mimimi_movement movement = {3, poses};
		
		mimimi_movement(animations->standing + i, model, &movement, allocator);
	}
	
	struct mimimi_pose_layer falling_layers[7] = {0};
	
	falling_layers[0].z_angle = coefficient;
	falling_layers[0].y_angle = 64 + 32 * coefficient;
	falling_layers[4].slant = 96 * coefficient;
	falling_layers[5].slant = 96 * coefficient;
	falling_layers[2].z_angle = coefficient;
	falling_layers[2].slant = -96;
	falling_layers[3].z_angle = coefficient;
	falling_layers[3].slant = 96;
	
	struct mimimi_pose falling_poses[] = {{x - 12 * coefficient, y - 12, 7, falling_layers}};
	struct mimimi_movement falling_movement = {1, falling_poses};
	mimimi_movement(animations->falling, model, &falling_movement, allocator);
	
	struct mimimi_pose_layer knocked_layers[1] = {0};
	
	knocked_layers[0].z_angle = coefficient;
	knocked_layers[0].y_angle = 64 + 32 * coefficient;
	
	struct mimimi_pose knocked_poses[] = {{x - 12 * coefficient, y - 20, 1, knocked_layers}};
	struct mimimi_movement knocked_movement = {1, knocked_poses};
	mimimi_movement(animations->knocked, model, &knocked_movement, allocator);
	
	int jumping_image_count = animations->jumping->count;
	for (int i = 0 ; i < jumping_image_count ; i++)
	{
		struct mimimi_pose_layer layers[6] = {0};
		
		int slant = 64 * i / jumping_image_count * coefficient;
		int y_angle = 64 + (20 + 20 * i / jumping_image_count) * coefficient;
		
		layers[0].y_angle = y_angle;
		layers[0].slant = -slant * 2;
		layers[1].slant = slant * 2;
		layers[4].slant = -slant / 2;
		layers[5].slant = -slant / 2;
		
		int count = sizeof layers / sizeof *layers;
		struct mimimi_pose pose = {x, y, count, layers};
		mimimi_pose(animations->jumping->images + i, model, &pose, allocator);
	}
	
	int landing_image_count = animations->landing->count;
	for (int i = 0 ; i < landing_image_count ; i++)
	{
		struct mimimi_pose_layer layers[6] = {0};
		
		int slant = 64 * i / landing_image_count * coefficient;
		int y_angle = 64 + (20 + 20 * i / landing_image_count) * coefficient;
		
		layers[0].y_angle = y_angle;
		layers[2].slant = -slant * 2;
		layers[3].slant = -slant * 2;
		layers[4].slant = slant * 2;
		layers[5].slant = slant * 2;
		
		int count = sizeof layers / sizeof *layers;
		struct mimimi_pose pose = {x, y, count, layers};
		mimimi_pose(animations->landing->images + i, model, &pose, allocator);
	}
}

struct mimimi_chapter_I_appearance *mimimi_chapter_I_appearance(struct mimimi_model *model, struct mimimi_allocator *allocator)
{
	struct mimimi_chapter_I_appearance *appearance = mimimi_allocate(allocator, sizeof *appearance);
	
	mimimi_chapter_I_animation_set(&appearance->left, allocator);
	mimimi_chapter_I_animation_set(&appearance->right, allocator);
	
	mimimi_chapter_I_appearance_animations(&appearance->left, model, 24, 37, 1, allocator);
	mimimi_chapter_I_appearance_animations(&appearance->right, model, 24, 37, -1, allocator);
	
	return appearance;
}

static void mimimi_chapter_I_allocate_video(struct mimimi_video *video, struct mimimi_animation *animation, struct mimimi_engine *engine, struct mimimi_allocator *allocator)
{
	video->count = animation->count;
	video->textures = mimimi_allocate(allocator, video->count * sizeof *video->textures);
	mimimi_video(video, animation, engine);
}

static void mimimi_chapter_I_video_set(struct mimimi_chapter_I_video_set *videos, struct mimimi_chapter_I_animation_set *animations, struct mimimi_engine *engine, struct mimimi_allocator *allocator)
{
	videos->standing_video_count = animations->standing_animation_count;
	videos->standing = mimimi_allocate(allocator, videos->standing_video_count * sizeof *videos->standing);
	videos->knocked = mimimi_allocate(allocator, sizeof *videos->knocked);
	videos->falling = mimimi_allocate(allocator, sizeof *videos->falling);
	videos->jumping = mimimi_allocate(allocator, sizeof *videos->jumping);
	videos->landing = mimimi_allocate(allocator, sizeof *videos->landing);
	
	for (int i = 0 ; i < videos->standing_video_count ; i++)
		mimimi_chapter_I_allocate_video(videos->standing + i, animations->standing + i, engine, allocator);
	mimimi_chapter_I_allocate_video(videos->knocked, animations->knocked, engine, allocator);
	mimimi_chapter_I_allocate_video(videos->falling, animations->falling, engine, allocator);
	mimimi_chapter_I_allocate_video(videos->jumping, animations->jumping, engine, allocator);
	mimimi_chapter_I_allocate_video(videos->landing, animations->landing, engine, allocator);
}

void mimimi_chapter_I_video_appearance(struct mimimi_chapter_I_video_appearance *video_appearance, struct mimimi_chapter_I_appearance *appearance, struct mimimi_engine *engine, struct mimimi_allocator *allocator)
{
	mimimi_chapter_I_video_set(&video_appearance->left, &appearance->left, engine, allocator);
	mimimi_chapter_I_video_set(&video_appearance->right, &appearance->right, engine, allocator);
}

struct mimimi_chapter_I_display
{
	struct mimimi_engine *engine;
	struct mimimi_position *camera;
	struct mimimi_sprite *sprite;
	struct mimimi_chapter_I_video_appearance *appearance;
	int x, y;
};

static void mimimi_chapter_I_start_display(struct mimimi_behavior *behavior, struct mimimi_behavior *compound, void *data)
{
	struct mimimi_chapter_I_display *display = data;
	struct mimimi_engine *engine = display->engine;
	struct mimimi_position *camera = display->camera;
	struct mimimi_sprite *sprite = display->sprite;
	struct mimimi_chapter_I_video_appearance *appearance = display->appearance;
	
	struct mimimi_position *position = sprite->position;
	struct mimimi_physics *physics = sprite->physics;
	struct mimimi_life *life = sprite->life;
	struct mimimi_walk *walk = sprite->walk;
	
	int direction = 1;
	unsigned char animation_time = 0;
	
	for (;;)
	{
		if (walk->direction != 0) direction = walk->direction;
		
		int dx;
		int dy = physics->dy;
		
		struct mimimi_chapter_I_video_set *videos;
		
		if (direction == 1)
		{
			dx = -physics->dx;
			videos = &appearance->left;
		}
		if (direction == 2)
		{
			dx = physics->dx;
			videos = &appearance->right;
		}
		
		if (dx > 31) dx = 31;
		if (dx < 0) dx = 0;
		if (dy < 0) dy = 0;
		
		struct mimimi_video *video;
		
		int x = -8 * display->x;
		int y = -8 * display->y;
		
		if (life->knocked_time == 0)
		{
			if (physics->airborne != 0)
			{
				struct mimimi_video *video = videos->jumping;
				if (physics->dy > 0) video = videos->landing;
				
				void *texture = video->textures[dx * videos->jumping->count / 32];
				mimimi_camera_stamp(engine, camera, position->x + x, position->y + y, 0, texture);
				(*behavior->behave)(behavior->data);
				continue;
			}
			
			video = videos->standing + dx * videos->standing_video_count / 32;
			animation_time += dx / 4;
		}
		else
		{
			if (physics->airborne == 0)
				video = videos->knocked,
				y += 128;
			else
				video = videos->falling,
				y += 32;
			
			animation_time += dy / 8;
		}
		
		// do not assume 'char' is eight bits
		animation_time &= 0xFF;
		
		void *texture = video->textures[animation_time * video->count / 256];
		mimimi_camera_stamp(engine, camera, position->x + x, position->y + y, 0, texture);
		
		(*behavior->behave)(behavior->data);
	}
}

void mimimi_chapter_I_display(struct mimimi_behavior *compound, struct mimimi_engine *engine, struct mimimi_position *camera, struct mimimi_sprite *sprite, struct mimimi_chapter_I_video_appearance *appearance, int x, int y, struct mimimi_allocator *allocator)
{
	struct mimimi_chapter_I_display *display = mimimi_allocate(allocator, sizeof *display);
	display->engine = engine;
	display->camera = camera;
	display->sprite = sprite;
	display->appearance = appearance;
	display->x = x;
	display->y = y;
	mimimi_coroutine(engine->coroutines, compound, &mimimi_chapter_I_start_display, display, allocator);
}
